package com.persagy.apm.alarmservice.group.management.service.impl;

import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.JSONArray;
import com.alibaba.fastjson.JSONObject;
import com.google.common.collect.Lists;
import com.persagy.apm.alarmservice.group.management.model.equip.vo.EquipItem;
import com.persagy.apm.alarmservice.group.management.model.equip.dto.QueryEquipDTO;
import com.persagy.apm.alarmservice.group.management.model.group.ConvertGroupTool;
import com.persagy.apm.alarmservice.group.management.model.group.vo.GroupInfoVO;
import com.persagy.apm.alarmservice.group.management.model.group.vo.EquipGroupInfoVO;
import com.persagy.apm.alarmservice.group.management.model.group.vo.GroupTreeVO;
import com.persagy.apm.alarmservice.group.management.service.GroupManagementService;
import com.persagy.apm.common.context.poems.PoemsContext;
import com.persagy.apm.common.model.vo.SimpleObjVO;
import com.persagy.apm.common.response.PageList;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.StringUtils;
import org.elasticsearch.action.get.GetRequest;
import org.elasticsearch.action.get.GetResponse;
import org.elasticsearch.action.search.SearchRequest;
import org.elasticsearch.action.search.SearchResponse;
import org.elasticsearch.client.RequestOptions;
import org.elasticsearch.client.RestHighLevelClient;
import org.elasticsearch.index.query.BoolQueryBuilder;
import org.elasticsearch.index.query.QueryBuilders;
import org.elasticsearch.index.query.TermQueryBuilder;
import org.elasticsearch.index.reindex.BulkByScrollResponse;
import org.elasticsearch.index.reindex.DeleteByQueryRequest;
import org.elasticsearch.rest.RestStatus;
import org.elasticsearch.search.SearchHit;
import org.elasticsearch.search.SearchHits;
import org.elasticsearch.search.builder.SearchSourceBuilder;
import org.elasticsearch.search.sort.SortOrder;
import org.springframework.beans.BeanUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.StandardCharsets;
import java.util.*;
import java.util.stream.Collectors;

import static cn.hutool.core.util.StrUtil.*;
import static com.persagy.apm.alarmservice.common.constants.ElasticStrConstant.INDEX_NAME;
import static org.apache.commons.lang3.math.NumberUtils.INTEGER_ONE;
import static org.apache.commons.lang3.math.NumberUtils.INTEGER_ZERO;

/**
 * 分组管理service
 * @author 易涛
 * @version 1.0
 * @date 2021/9/8 14:18
 */
@Slf4j
@Service
public class GroupManagementServiceImpl implements GroupManagementService {

    @Autowired
    private RestHighLevelClient client;

    @Value("${elasticsearch.maxSearchCount}")
    private int maxSearchCount;

    @Override
    public PageList<EquipItem> equipsQuery(QueryEquipDTO managementDTO) {
        List<EquipItem> equipListItemList = new ArrayList<>();
        BoolQueryBuilder queryBuilder = QueryBuilders.boolQuery();
        queryBuilder.must(QueryBuilders.matchQuery("group_code", PoemsContext.getContext().getGroupCode()));
        queryBuilder.must(QueryBuilders.matchQuery("class_code", managementDTO.getClassCode()));
        if (StringUtils.isNotBlank(managementDTO.getCondition())) {
            queryBuilder.must(QueryBuilders.matchPhraseQuery("infos", managementDTO.getCondition()));
        }
        //新增根据项目id查询设备
        if (StringUtils.isNotEmpty(managementDTO.getProjectId())) {
            queryBuilder.must(QueryBuilders.matchQuery("project_id", managementDTO.getProjectId()));
        }
        //模糊查询使用
        if (StringUtils.isNotEmpty(managementDTO.getLocalName())) {
            queryBuilder.must(QueryBuilders.matchQuery("local_name", managementDTO.getLocalName())
                    .analyzer("keyword"));
        }
        queryBuilder.must(QueryBuilders.termQuery("valid", INTEGER_ONE));
        SearchSourceBuilder sourceBuilder = new SearchSourceBuilder();
        sourceBuilder.query(queryBuilder);
        sourceBuilder.from((Math.max(managementDTO.getPage(), INTEGER_ONE) - INTEGER_ONE)
                * managementDTO.getSize());
        sourceBuilder.size(managementDTO.getSize());
        sourceBuilder.sort("create_time", SortOrder.DESC);
        SearchRequest rq = new SearchRequest(INDEX_NAME);
        rq.source(sourceBuilder);
        try {
            SearchResponse searchResponse = client.search(rq, RequestOptions.DEFAULT);
            // 根据状态和数据条数验证是否返回了数据
            if (RestStatus.OK.equals(searchResponse.status()) && searchResponse.getHits().getTotalHits().value > 0) {
                SearchHits hits = searchResponse.getHits();
                for (SearchHit hit : hits) {
                    // 将 JSON 转换成对象
                    EquipItem managementVo = JSON.parseObject(hit.getSourceAsString(), EquipItem.class);
                    equipListItemList.add(managementVo);
                }
            }
            return new PageList<>(equipListItemList, searchResponse.getHits().getTotalHits().value);
        } catch (IOException e) {
            throw new RuntimeException("es查询出错", e);
        }
    }

    @Override
    public EquipItem equipGet(String id) {
        String equipStr = findEquipById(id);
        if (StringUtils.isEmpty(equipStr)) {
            throw new RuntimeException("id为" + id + "的设备不存在");
        }
        return JSONObject.parseObject(equipStr, EquipItem.class);
    }

    /**
     * 根据id查询设备-常用方法
     * @param id 设备id
     * @author 易涛
     * @date 2021/9/10 10:05
     */
    private String findEquipById(String id) {
        GetRequest getRequest = new GetRequest(INDEX_NAME, id);
        try {
            GetResponse getResponse = client.get(getRequest, RequestOptions.DEFAULT);
            if (getResponse.isExists()) {
                return getResponse.getSourceAsString();
            }
        } catch (IOException e) {
            throw new RuntimeException("es查询出错", e);
        }
        return null;
    }

    /**
     * 查询设备
     *
     * @param ids 设备id列表
     * @return 设备列表
     * @author lixing
     * @version V1.0 2021/9/20 1:15 下午
     */
    private List<EquipItem> findEquipsByIds(List<String> ids) {
        if (CollectionUtils.isEmpty(ids) || ids.size() > 100) {
            throw new RuntimeException("批量查询的数据过短或过长");
        }
        List<EquipItem> managementList = new ArrayList<>();
        BoolQueryBuilder queryBuilder = QueryBuilders.boolQuery();
        queryBuilder.must(QueryBuilders.termsQuery("id.keyword", ids));
        SearchSourceBuilder sourceBuilder = new SearchSourceBuilder();
        sourceBuilder.query(queryBuilder);
        sourceBuilder.from(INTEGER_ZERO);
        sourceBuilder.size(ids.size());
        sourceBuilder.sort("create_time", SortOrder.DESC);
        SearchRequest rq = new SearchRequest(INDEX_NAME);
        rq.source(sourceBuilder);
        try {
            SearchResponse searchResponse = client.search(rq, RequestOptions.DEFAULT);
            // 根据状态和数据条数验证是否返回了数据
            if (RestStatus.OK.equals(searchResponse.status()) && searchResponse.getHits().getTotalHits().value > 0) {
                SearchHits hits = searchResponse.getHits();
                for (SearchHit hit : hits) {
                    // 将 JSON 转换成对象
                    EquipItem management = JSON.parseObject(hit.getSourceAsString(), EquipItem.class);
                    managementList.add(management);
                }
            }
        } catch (IOException e) {
            throw new RuntimeException("es查询出错", e);
        }
        return managementList;
    }

    @Override
    public Map<String, EquipItem> queryEquipsByIds(List<String> ids) {
        if (CollectionUtils.isEmpty(ids)) {
            return new HashMap<>();
        }
        Map<String, EquipItem> resultMap = new HashMap<>(ids.size());
        List<List<String>> partitions = Lists.partition(ids, 100);
        for (List<String> partition : partitions) {
            List<EquipItem> equipsByIds = findEquipsByIds(ids);
            if (CollectionUtils.isEmpty(equipsByIds)) {
                continue;
            }
            Map<String, EquipItem> partitionMap = equipsByIds.stream().collect(Collectors.toMap(
                    EquipItem::getId, equip -> equip, (k1, k2) -> k1
            ));
            resultMap.putAll(partitionMap);
        }
        return resultMap;
    }

    @Override
    public List<EquipGroupInfoVO> equipGroupQuery(List<String> ids) {
        List<EquipGroupInfoVO> equipGroupInfoVOList = Lists.newArrayList();
        List<EquipItem> equipItemList = findEquipsByIds(ids);
        for(EquipItem equipItem: equipItemList) {
            if(StringUtils.isEmpty(equipItem.getInfos())){
                return equipGroupInfoVOList;
            }
            JSONObject jsonObject = JSONObject.parseObject(equipItem.getInfos());
            Map<String, String> equipFuncMap = new HashMap<>();
            if (jsonObject != null) {
                for (String key : jsonObject.keySet()) {
                    String funcInfo = key + ":" + jsonObject.get(key).toString();
                    equipFuncMap.put(funcInfo, funcInfo);
                }
            }
            //获取分组树
            List<GroupTreeVO> groupTreeVOList = getGroupTree();
            EquipGroupInfoVO equipGroupInfoVO = new EquipGroupInfoVO();
            equipGroupInfoVO.setEquipId(equipItem.getId());
            equipGroupInfoVO.setClassCode(equipItem.getClassCode());
            if(CollectionUtils.isEmpty(groupTreeVOList)){
                return equipGroupInfoVOList;
            }
            equipGroupInfoVOList.add(cascadeSearchGroupInfo(groupTreeVOList,equipGroupInfoVO,equipFuncMap));
        }
        return equipGroupInfoVOList;
    }

    @Override
    public Map<String, SimpleObjVO> queryObjGroup(List<String> objIds) {
        if (CollectionUtils.isEmpty(objIds)) {
            return new HashMap<>();
        }
        Map<String, SimpleObjVO> resultMap = new HashMap<>(objIds.size());

        // 每次查询100条
        List<List<String>> partitions = Lists.partition(objIds, 100);
        for (List<String> partition : partitions) {
            List<EquipGroupInfoVO> groupTreeInfoList = equipGroupQuery(partition);
            if (CollectionUtils.isEmpty(groupTreeInfoList)) {
                continue;
            }
            Map<String, SimpleObjVO> partitionMap = groupTreeInfoList.stream().collect(Collectors.toMap(
                    EquipGroupInfoVO::getEquipId,
                    equipGroupInfoVO -> new SimpleObjVO(equipGroupInfoVO.getId(), equipGroupInfoVO.getName()),
                    (k1, k2) -> k1
            ));
            resultMap.putAll(partitionMap);
        }

        return resultMap;
    }

    /**
     * 递归查询设备的分组信息
     * 遍历分组树，如果设备的信息点中包含分组的分组条件，同时设备的设备类编码与分组设备类编码匹配时，代表设备属于这个分组
     *
     * @param groupTreeVOList 分组树节点列表
     * @param equipGroupInfoVO 设备和所在分组信息(初始化时包含设备id和设备类编码)
     * @param equipFuncMap "设备信息点:信息点取值" -> "设备信息点:信息点取值"，
     * @return 设备和所在分组信息（补充完成分组信息）
     * @author 易涛
     * @date 2021/9/10 11:24
     */
    private EquipGroupInfoVO cascadeSearchGroupInfo(
            List<GroupTreeVO> groupTreeVOList, EquipGroupInfoVO equipGroupInfoVO,
            Map<String, String> equipFuncMap) {
        if (StringUtils.isBlank(equipGroupInfoVO.getClassCode())) {
            throw new IllegalArgumentException("要查询的设备没有设备类信息");
        }
        for (GroupTreeVO groupTreeVO : groupTreeVOList) {
            if (!equipGroupInfoVO.getClassCode().equals(groupTreeVO.getClassCode())) {
                continue;
            }
            // 分组条件 排除掉配置的时候存在不小心放入空格的情况
            String groupCondition = groupTreeVO.getCondition().replace(" ", "");
            // 分组条件为空或设备信息点中包含分组条件
            if (StringUtils.isBlank(groupCondition) || equipFuncMap.containsKey(groupCondition)) {
                BeanUtils.copyProperties(groupTreeVO, equipGroupInfoVO);
                return equipGroupInfoVO;
            }

            if (CollectionUtils.isNotEmpty(groupTreeVO.getChildren())) {
                cascadeSearchGroupInfo(groupTreeVO.getChildren(), equipGroupInfoVO, equipFuncMap);
            }
        }
        return equipGroupInfoVO;
    }

    /**
     * 获取分组树
     *
     * @return java.util.List<com.persagy.apm.alarmservice.group.management.model.vo.GroupTreeVO>
     * @author 易涛
     * @date 2021/9/8 14:44
     */
    public static List<GroupTreeVO> getGroupTree() {
        InputStream inputStream = GroupManagementServiceImpl.class.getClassLoader()
                .getResourceAsStream( "json" + File.separator + "GroupTree.json");
        if (inputStream == null) {
            return null;
        }
        try {
            String str = IOUtils.toString(inputStream, StandardCharsets.UTF_8);
            return JSONArray.parseArray(str, GroupTreeVO.class);
        } catch (IOException ioe) {
            throw new RuntimeException("json文件有误", ioe);
        }
    }

    @Override
    public void batchDelete() {
        DeleteByQueryRequest request = new DeleteByQueryRequest(INDEX_NAME);
        //逻辑删除的数据直接从index中删除掉
        request.setQuery(new TermQueryBuilder("valid", INTEGER_ZERO));
        //每次最多删除一万条数据，如果不加这个则处理所有的文档
        request.setMaxDocs(10000);
        try {
            BulkByScrollResponse resp = client.deleteByQuery(request, RequestOptions.DEFAULT);
            log.info("删除返回的数据为---{}", resp.toString());
        } catch (IOException e) {
            throw new RuntimeException("删除elasticSearch失败", e);
        }
    }

    @Override
    public List<EquipItem> getEquipList(String condition, String classCode) {
        QueryEquipDTO managementDTO = new QueryEquipDTO();
        managementDTO.setClassCode(classCode);
        managementDTO.setCondition(condition);
        return getEquipList(managementDTO);
    }

    @Override
    public List<EquipItem> getEquipList(QueryEquipDTO queryEquipDTO) {
        queryEquipDTO.setPage(INTEGER_ONE);
        queryEquipDTO.setSize(maxSearchCount);
        PageList<EquipItem> groupManagementVoPageList = this.equipsQuery(queryEquipDTO);
        return groupManagementVoPageList.getRecords();
    }

    @Override
    public GroupInfoVO getGroup(String id) {
        return queryById(getGroupTree(), new GroupInfoVO(), id);
    }

    /**
     * 递归匹配符合条件的分组对象
     *
     * @param groupTreeVOList
     * @param id
     * @return com.persagy.apm.alarmservice.group.management.model.vo.GroupTreeVO
     * @author 易涛
     * @date 2021/9/8 14:50
     */
    private GroupInfoVO queryById(List<GroupTreeVO> groupTreeVOList, GroupInfoVO groupInfoVO, String id) {
        if (CollectionUtils.isEmpty(groupTreeVOList)) {
            return null;
        }
        for (GroupTreeVO groupTreeVO : groupTreeVOList) {
            if (StringUtils.isNotEmpty(groupInfoVO.getId())) {
                return groupInfoVO;
            }
            if (Objects.equals(groupTreeVO.getId(), id)) {
                BeanUtils.copyProperties(groupTreeVO, groupInfoVO);
                return groupInfoVO;
            }
            if (CollectionUtils.isNotEmpty(groupTreeVO.getChildren())) {
                queryById(groupTreeVO.getChildren(), groupInfoVO, id);
            }
        }
        return groupInfoVO;
    }

    @Override
    public List<GroupInfoVO> queryAllLeafGroups() {
        // 遍历分组树，获取最末级节点
        List<GroupTreeVO> groupTree = getGroupTree();

        List<GroupInfoVO> result = new ArrayList<>();
        cascadeGetLeafGroup(groupTree, result);
        return result;
    }

    /**
     * 递归查询分组树的叶子节点
     *
     * @param groupTree 分组树节点列表
     * @param result 保存叶子节点的结果集
     * @author lixing
     * @version V1.0 2021/9/19 1:37 下午
     */
    private void cascadeGetLeafGroup(List<GroupTreeVO> groupTree, List<GroupInfoVO> result) {
        if (CollectionUtils.isEmpty(groupTree)) {
            return;
        }
        for (GroupTreeVO groupTreeVO : groupTree) {
            if (CollectionUtils.isEmpty(groupTreeVO.getChildren())) {
                result.add(ConvertGroupTool.INSTANCE.convert2DetailDTO(groupTreeVO));
            } else {
                cascadeGetLeafGroup(groupTreeVO.getChildren(), result);
            }
        }
    }
}
